// ******************************************************************
// *
// *    .,-:::::    .,::      .::::::::.    .,::      .:
// *  ,;;;'````'    `;;;,  .,;;  ;;;'';;'   `;;;,  .,;; 
// *  [[[             '[[,,[['   [[[__[[\.    '[[,,[['  
// *  $$$              Y$$$P     $$""""Y$$     Y$$$P    
// *  `88bo,__,o,    oP"``"Yo,  _88o,,od8P   oP"``"Yo,  
// *    "YUMMMMMP",m"       "Mm,""YUMMMP" ,m"       "Mm,
// *
// *   Cxbx->Win32->CxbxKrnl->EmuShared.cpp
// *
// *  This file is part of the Cxbx project.
// *
// *  Cxbx and Cxbe are free software; you can redistribute them
// *  and/or modify them under the terms of the GNU General Public
// *  License as published by the Free Software Foundation; either
// *  version 2 of the license, or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2003 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************
///#define _CXBXKRNL_INTERNAL
///#define _XBOXKRNL_DEFEXTRN_

// ******************************************************************
// *
// *    .,-:::::    .,::      .::::::::.    .,::      .:
// *  ,;;;'````'    `;;;,  .,;;  ;;;'';;'   `;;;,  .,;; 
// *  [[[             '[[,,[['   [[[__[[\.    '[[,,[['  
// *  $$$              Y$$$P     $$""""Y$$     Y$$$P    
// *  `88bo,__,o,    oP"``"Yo,  _88o,,od8P   oP"``"Yo,  
// *    "YUMMMMMP",m"       "Mm,""YUMMMP" ,m"       "Mm,
// *
// *   Cxbx->Win32->CxbxKrnl->EmuShared.h
// *
// *  This file is part of the Cxbx project.
// *
// *  Cxbx and Cxbe are free software; you can redistribute them
// *  and/or modify them under the terms of the GNU General Public
// *  License as published by the Free Software Foundation; either
// *  version 2 of the license, or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2003 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************



// ******************************************************************
// * EmuShared : Shared memory
// ******************************************************************
public class EmuShared extends Mutex
{
		// ******************************************************************
		// * Constructor / Deconstructor
		// ******************************************************************

		// ******************************************************************
		// * Constructor
		// ******************************************************************
		public CXBXKRNL_API EmuShared()
		{
			m_XBController.Load("Software\\Cxbx\\XBController");
			m_XBVideo.Load("Software\\Cxbx\\XBVideo");
		}

		// ******************************************************************
		// * Deconstructor
		// ******************************************************************
		public final CXBXKRNL_API~EmuShared()
		{
			m_XBController.Save("Software\\Cxbx\\XBController");
			m_XBVideo.Save("Software\\Cxbx\\XBVideo");
		}

		// ******************************************************************
		// * Each process needs to call this to initialize shared memory
		// ******************************************************************

		// ******************************************************************
		// * func: EmuShared::EmuSharedInit
		// ******************************************************************
		public static CXBXKRNL_API void Init()
		{
			// ******************************************************************
			// * Ensure initialization only occurs once
			// ******************************************************************
			boolean init = true;

			// ******************************************************************
			// * Prevent multiple initializations
			// ******************************************************************
			if (GlobalMembers.hMapObject != null)
			{
				return;
			}

			// ******************************************************************
			// * Create the shared memory "file"
			// ******************************************************************
			{
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
				GlobalMembers.hMapObject = CreateFileMapping(INVALID_HANDLE_VALUE, null, PAGE_READWRITE, 0, sizeof(GlobalMembers.EmuShared), "Local\\EmuShared"); // name of map object -  size: low 32 bits -  size: high 32 bits -  read/write access -  default security attributes -  Paging file

				if (GlobalMembers.hMapObject == null)
				{
					CxbxKrnlCleanup("Could not map shared memory!");
				}

				if (GetLastError() == ERROR_ALREADY_EXISTS)
				{
					init = false;
				}
			}

			// ******************************************************************
			// * Memory map this file
			// ******************************************************************
			{
				g_EmuShared = (GlobalMembers.EmuShared)MapViewOfFile(GlobalMembers.hMapObject, FILE_MAP_WRITE, 0, 0, 0); // default: map entire file -  low offset:   beginning -  high offset:  map from -  read/write access -  object to map view of

				if (g_EmuShared == null)
				{
					CxbxKrnlCleanup("Could not map view of shared memory!");
				}
			}

			// ******************************************************************
			// * Executed only on first initialization of shared memory
			// ******************************************************************
			if (init)
			{
				g_EmuShared.EmuShared.EmuShared();
			}

			g_EmuSharedRefCount++;
		}

		// ******************************************************************
		// * Each process needs to call this to cleanup shared memory
		// ******************************************************************

		// ******************************************************************
		// * func: EmuSharedCleanup
		// ******************************************************************
		public static CXBXKRNL_API void Cleanup()
		{
			g_EmuSharedRefCount--;

			if (g_EmuSharedRefCount == 0)
			{
				g_EmuShared.EmuShared.~new GlobalMembers.EmuShared();

				UnmapViewOfFile(g_EmuShared);
			}
		}

		// ******************************************************************
		// * Xbox Video Accessors
		// ******************************************************************
		public final CXBXKRNL_API void GetXBVideo(XBVideo video)
		{
			Lock();
//C++ TO JAVA CONVERTER TODO TASK: The memory management function 'memcpy' has no equivalent in Java:
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			memcpy(video, m_XBVideo, sizeof(XBVideo));
			Unlock();
		}
		public final CXBXKRNL_API void SetXBVideo(XBVideo video)
		{
			Lock();
//C++ TO JAVA CONVERTER TODO TASK: The memory management function 'memcpy' has no equivalent in Java:
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			memcpy(m_XBVideo, video, sizeof(XBVideo));
			Unlock();
		}

		// ******************************************************************
		// * Xbox Controller Accessors
		// ******************************************************************
		public final CXBXKRNL_API void GetXBController(XBController ctrl)
		{
			Lock();
//C++ TO JAVA CONVERTER TODO TASK: The memory management function 'memcpy' has no equivalent in Java:
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			memcpy(ctrl, m_XBController, sizeof(XBController));
			Unlock();
		}
		public final CXBXKRNL_API void SetXBController(XBController ctrl)
		{
			Lock();
//C++ TO JAVA CONVERTER TODO TASK: The memory management function 'memcpy' has no equivalent in Java:
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			memcpy(m_XBController, ctrl, sizeof(XBController));
			Unlock();
		}

		// ******************************************************************
		// * Xbe Path Accessors
		// ******************************************************************
		public final CXBXKRNL_API void GetXbePath(tangible.RefObject<String> path)
		{
			Lock();
			path.argValue = m_XbePath;
			Unlock();
		}
		public final CXBXKRNL_API void SetXbePath(String path)
		{
			Lock();
			m_XbePath = path;
			Unlock();
		}

		// ******************************************************************
		// * Shared configuration
		// ******************************************************************
		private XBController m_XBController = new XBController();
		private XBVideo m_XBVideo = new XBVideo();
		private String m_XbePath = new String(new char[260]);
}

public class GlobalMembers
{
	// ******************************************************************
	// * Exported Global Shared Memory Pointer
	// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: 'extern' variable declarations are not required in Java:
	//extern CXBXKRNL_API EmuShared *g_EmuShared;
//C++ TO JAVA CONVERTER NOTE: 'extern' variable declarations are not required in Java:
	//extern CXBXKRNL_API int g_EmuSharedRefCount;



	//C++ TO JAVA CONVERTER TODO TASK: There is no preprocessor in Java:
	///#undef FIELD_OFFSET

	// ******************************************************************
	// * exported globals
	// ******************************************************************
	public static CXBXKRNL_API EmuShared * g_EmuShared = null;
	public static CXBXKRNL_API int g_EmuSharedRefCount = 0;

	// ******************************************************************
	// * static/global
	// ******************************************************************
	public static HANDLE hMapObject = null;
}

final class DefineConstants
{
//C++ TO JAVA CONVERTER TODO TASK: The following #define constant was defined in different ways:
	public static final String _CXBX_VERSION = "0.8.0";
	public static final String _CXBX_VERSION = "0.8.0-Trace";
	public static final int MAXIMUM_XBOX_THREADS = 256;
	public static final int XINPUT_SETSTATE_LATENCY = 4;
	public static final int XINPUT_SETSTATE_SLOTS = 16;
	public static final int XINPUT_HANDLE_SLOTS = 4;
	public static final int XINPUT_DEVSUBTYPE_GC_GAMEPAD = 0x01;
	public static final int XINPUT_DEVSUBTYPE_GC_GAMEPAD_ALT = 0x02;
	public static final int XINPUT_DEVSUBTYPE_GC_WHEEL = 0x10;
	public static final int XINPUT_DEVSUBTYPE_GC_ARCADE_STICK = 0x20;
	public static final int XINPUT_DEVSUBTYPE_GC_DIGITAL_ARCADE_STICK = 0x21;
	public static final int XINPUT_DEVSUBTYPE_GC_FLIGHT_STICK = 0x30;
	public static final int XINPUT_DEVSUBTYPE_GC_SNOWBOARD = 0x40;
	public static final int XCALCSIG_SIGNATURE_SIZE = 20;
	public static final int XCALCSIG_FLAG_NON_ROAMABLE = 0x00000001;
	public static final int DIRECT3D_VERSION = 0x0800;
	public static final int X_PIXELSHADER_FAKE_HANDLE = 0xDEADBEEF;
	public static final int X_D3DCOMMON_REFCOUNT_MASK = 0x0000FFFF;
	public static final int X_D3DCOMMON_TYPE_MASK = 0x00070000;
	public static final int X_D3DCOMMON_TYPE_SHIFT = 16;
	public static final int X_D3DCOMMON_TYPE_VERTEXBUFFER = 0x00000000;
	public static final int X_D3DCOMMON_TYPE_INDEXBUFFER = 0x00010000;
	public static final int X_D3DCOMMON_TYPE_PUSHBUFFER = 0x00020000;
	public static final int X_D3DCOMMON_TYPE_PALETTE = 0x00030000;
	public static final int X_D3DCOMMON_TYPE_TEXTURE = 0x00040000;
	public static final int X_D3DCOMMON_TYPE_SURFACE = 0x00050000;
	public static final int X_D3DCOMMON_TYPE_FIXUP = 0x00060000;
	public static final int X_D3DCOMMON_INTREFCOUNT_MASK = 0x00780000;
	public static final int X_D3DCOMMON_INTREFCOUNT_SHIFT = 19;
	public static final int X_D3DCOMMON_D3DCREATED = 0x01000000;
	public static final int X_D3DCOMMON_ISLOCKED = 0x02000010; // Surface is currently locked (potential unswizzle candidate)
	public static final int X_D3DCOMMON_UNUSED_MASK = 0xFE000000;
	public static final int X_D3DCOMMON_UNUSED_SHIFT = 25;
	public static final int X_D3DRESOURCE_DATA_FLAG_SPECIAL = 0xFFFF0000;
	public static final int X_D3DRESOURCE_DATA_FLAG_SURFACE = 0x00000001; // Backbuffer surface, etc
	public static final int X_D3DRESOURCE_DATA_FLAG_YUVSURF = 0x00000002; // YUV memory surface
	public static final int X_D3DRESOURCE_DATA_FLAG_D3DREND = 0x00000004; // D3D Render Target
	public static final int X_D3DRESOURCE_DATA_FLAG_D3DSTEN = 0x00000008; // D3D Stencil Surface
	public static final int X_D3DRESOURCE_LOCK_FLAG_NOSIZE = 0xEFFFFFFF;
	public static final int X_D3DFORMAT_RESERVED1_MASK = 0x00000003; // Must be zero
	public static final int X_D3DFORMAT_DMACHANNEL_MASK = 0x00000003;
	public static final int X_D3DFORMAT_DMACHANNEL_A = 0x00000001; // DMA channel A - the default for all system memory
	public static final int X_D3DFORMAT_DMACHANNEL_B = 0x00000002; // DMA channel B - unused
	public static final int X_D3DFORMAT_CUBEMAP = 0x00000004; // Set if the texture if a cube map
	public static final int X_D3DFORMAT_BORDERSOURCE_COLOR = 0x00000008;
	public static final int X_D3DFORMAT_DIMENSION_MASK = 0x000000F0; // # of dimensions
	public static final int X_D3DFORMAT_DIMENSION_SHIFT = 4;
	public static final int X_D3DFORMAT_FORMAT_MASK = 0x0000FF00;
	public static final int X_D3DFORMAT_FORMAT_SHIFT = 8;
	public static final int X_D3DFORMAT_MIPMAP_MASK = 0x000F0000;
	public static final int X_D3DFORMAT_MIPMAP_SHIFT = 16;
	public static final int X_D3DFORMAT_USIZE_MASK = 0x00F00000; // Log 2 of the U size of the base texture
	public static final int X_D3DFORMAT_USIZE_SHIFT = 20;
	public static final int X_D3DFORMAT_VSIZE_MASK = 0x0F000000; // Log 2 of the V size of the base texture
	public static final int X_D3DFORMAT_VSIZE_SHIFT = 24;
	public static final int X_D3DFORMAT_PSIZE_MASK = 0xF0000000; // Log 2 of the P size of the base texture
	public static final int X_D3DFORMAT_PSIZE_SHIFT = 28;
	public static final int X_D3DSIZE_WIDTH_MASK = 0x00000FFF; // Width  (Texels - 1)
	public static final int X_D3DSIZE_HEIGHT_MASK = 0x00FFF000; // Height (Texels - 1)
	public static final int X_D3DSIZE_HEIGHT_SHIFT = 12;
	public static final int X_D3DSIZE_PITCH_MASK = 0xFF000000; // Pitch / 64 - 1
	public static final int X_D3DSIZE_PITCH_SHIFT = 24;
	public static final int X_D3DRS_UNK = 0x7fffffff;
	public static final int X_D3DTSS_UNK = 0x7fffffff;
	public static final int X_VSCM_96 = 0;
	public static final int X_VSCM_192 = 1;
	public static final int X_VSCM_192FIXEDPIPELINE = 2;
	public static final int X_VSCM_NONERESERVED = 16;
	public static final int X_VST_NORMAL = 1;
	public static final int X_VST_READWRITE = 2;
	public static final int X_VST_STATE = 3;
	public static final int DIRECTDRAW_VERSION = 0x0700;
	public static final int X_D3DRSSE_UNK = 0x7fffffff;
	public static final int MAX_NBR_STREAMS = 16;
	public static final int VSH_INSTRUCTION_SIZE = 4;
	public static final int DIRECTINPUT_VERSION = 0x0800;
	public static final int XINPUT_GAMEPAD_A = 0;
	public static final int XINPUT_GAMEPAD_B = 1;
	public static final int XINPUT_GAMEPAD_X = 2;
	public static final int XINPUT_GAMEPAD_Y = 3;
	public static final int XINPUT_GAMEPAD_BLACK = 4;
	public static final int XINPUT_GAMEPAD_WHITE = 5;
	public static final int XINPUT_GAMEPAD_LEFT_TRIGGER = 6;
	public static final int XINPUT_GAMEPAD_RIGHT_TRIGGER = 7;
	public static final int XINPUT_GAMEPAD_DPAD_UP = 0x00000001;
	public static final int XINPUT_GAMEPAD_DPAD_DOWN = 0x00000002;
	public static final int XINPUT_GAMEPAD_DPAD_LEFT = 0x00000004;
	public static final int XINPUT_GAMEPAD_DPAD_RIGHT = 0x00000008;
	public static final int XINPUT_GAMEPAD_START = 0x00000010;
	public static final int XINPUT_GAMEPAD_BACK = 0x00000020;
	public static final int XINPUT_GAMEPAD_LEFT_THUMB = 0x00000040;
	public static final int XINPUT_GAMEPAD_RIGHT_THUMB = 0x00000080;
	public static final int DSB_FLAG_ADPCM = 0x00000001;
	public static final int DETECT_SENSITIVITY_JOYSTICK = 25000;
	public static final int DETECT_SENSITIVITY_BUTTON = 0;
	public static final int DETECT_SENSITIVITY_MOUSE = 5;
	public static final int DETECT_SENSITIVITY_POV = 50000;
}

package tangible;

//----------------------------------------------------------------------------------------
//	Copyright © 2006 - 2016 Tangible Software Solutions Inc.
//	This class can be used by anyone provided that the copyright notice remains intact.
//
//	This class is used to simulate the ability to pass arguments by reference in Java.
//----------------------------------------------------------------------------------------
public final class RefObject<T>
{
	public T argValue;
	public RefObject(T refArg)
	{
		argValue = refArg;
	}
}