import javax.swing.*;

// ******************************************************************
// *
// *    .,-:::::    .,::      .::::::::.    .,::      .:
// *  ,;;;'````'    `;;;,  .,;;  ;;;'';;'   `;;;,  .,;; 
// *  [[[             '[[,,[['   [[[__[[\.    '[[,,[['  
// *  $$$              Y$$$P     $$""""Y$$     Y$$$P    
// *  `88bo,__,o,    oP"``"Yo,  _88o,,od8P   oP"``"Yo,  
// *    "YUMMMMMP",m"       "Mm,""YUMMMP" ,m"       "Mm,
// *
// *   Cxbx->Win32->XBController.cpp
// *
// *  This file is part of the Cxbx project.
// *
// *  Cxbx and Cxbe are free software; you can redistribute them
// *  and/or modify them under the terms of the GNU General Public
// *  License as published by the Free Software Foundation; either
// *  version 2 of the license, or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2003 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************
// ******************************************************************
// *
// *    .,-:::::    .,::      .::::::::.    .,::      .:
// *  ,;;;'````'    `;;;,  .,;;  ;;;'';;'   `;;;,  .,;; 
// *  [[[             '[[,,[['   [[[__[[\.    '[[,,[['  
// *  $$$              Y$$$P     $$""""Y$$     Y$$$P    
// *  `88bo,__,o,    oP"``"Yo,  _88o,,od8P   oP"``"Yo,  
// *    "YUMMMMMP",m"       "Mm,""YUMMMP" ,m"       "Mm,
// *
// *   Cxbx->Win32->Cxbx->XBController.h
// *
// *  This file is part of the Cxbx project.
// *
// *  Cxbx and Cxbe are free software; you can redistribute them
// *  and/or modify them under the terms of the GNU General Public
// *  License as published by the Free Software Foundation; either
// *  version 2 of the license, or (at your option) any later version.
// *
// *  This program is distributed in the hope that it will be useful,
// *  but WITHOUT ANY WARRANTY; without even the implied warranty of
// *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// *  GNU General Public License for more details.
// *
// *  You should have recieved a copy of the GNU General Public License
// *  along with this program; see the file COPYING.
// *  If not, write to the Free Software Foundation, Inc.,
// *  59 Temple Place - Suite 330, Bostom, MA 02111-1307, USA.
// *
// *  (c) 2002-2003 Aaron Robinson <caustik@caustik.com>
// *
// *  All rights reserved
// *
// ******************************************************************




// ******************************************************************
// * Xbox Controller Object IDs
// ******************************************************************
public enum XBCtrlObject
{
	// ******************************************************************
	// * Analog Axis
	// ******************************************************************
	XBCTRL_OBJECT_LTHUMBPOSX(0),
	XBCTRL_OBJECT_LTHUMBNEGX(1),
	XBCTRL_OBJECT_LTHUMBPOSY(2),
	XBCTRL_OBJECT_LTHUMBNEGY(3),
	XBCTRL_OBJECT_RTHUMBPOSX(4),
	XBCTRL_OBJECT_RTHUMBNEGX(5),
	XBCTRL_OBJECT_RTHUMBPOSY(6),
	XBCTRL_OBJECT_RTHUMBNEGY(7),
	// ******************************************************************
	// * Analog Buttons
	// ******************************************************************
	XBCTRL_OBJECT_A(8),
	XBCTRL_OBJECT_B(9),
	XBCTRL_OBJECT_X(10),
	XBCTRL_OBJECT_Y(11),
	XBCTRL_OBJECT_BLACK(12),
	XBCTRL_OBJECT_WHITE(13),
	XBCTRL_OBJECT_LTRIGGER(14),
	XBCTRL_OBJECT_RTRIGGER(15),
	// ******************************************************************
	// * Digital Buttons
	// ******************************************************************
	XBCTRL_OBJECT_DPADUP(16),
	XBCTRL_OBJECT_DPADDOWN(17),
	XBCTRL_OBJECT_DPADLEFT(18),
	XBCTRL_OBJECT_DPADRIGHT(19),
	XBCTRL_OBJECT_BACK(20),
	XBCTRL_OBJECT_START(21),
	XBCTRL_OBJECT_LTHUMB(22),
	XBCTRL_OBJECT_RTHUMB(23),
	// ******************************************************************
	// * Total number of components
	// ******************************************************************
	XBCTRL_OBJECT_COUNT(24);

	private int intValue;
	private static java.util.HashMap<Integer, XBCtrlObject> mappings;
	private static java.util.HashMap<Integer, XBCtrlObject> getMappings()
	{
		if (mappings == null)
		{
			synchronized (XBCtrlObject.class)
			{
				if (mappings == null)
				{
					mappings = new java.util.HashMap<Integer, XBCtrlObject>();
				}
			}
		}
		return mappings;
	}

	private XBCtrlObject(int value)
	{
		intValue = value;
		getMappings().put(value, this);
	}

	public int getValue()
	{
		return intValue;
	}

	public static XBCtrlObject forValue(int value)
	{
		return getMappings().get(value);
	}
}

// ******************************************************************
// * Maximum number of devices allowed
// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define XBCTRL_MAX_DEVICES XBCTRL_OBJECT_COUNT

// ******************************************************************
// * Xbox Controller Object Config
// ******************************************************************
public class XBCtrlObjectCfg
{
	public int dwDevice; // offset into m_InputDevice
	public int dwInfo; // extended information, depending on dwFlags
	public int dwFlags; // flags explaining the data format
}

// ******************************************************************
// * class: XBController
// ******************************************************************
public class XBController extends Error
{
		// ******************************************************************
		// * Initialization
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::XBController
		// ******************************************************************
		public XBController()
		{
			m_CurrentState = XBCtrlState.XBCTRL_STATE_NONE;

			int v = 0;

			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				m_DeviceName[v][0] = (byte)'\0';

				m_InputDevice[v].m_Device = null;
				m_InputDevice[v].m_Flags = 0;
			}

			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				m_ObjectConfig[v].dwDevice = -1;
				m_ObjectConfig[v].dwInfo = -1;
				m_ObjectConfig[v].dwFlags = 0;
			}

			m_pDirectInput8 = null;

			m_dwInputDeviceCount = 0;
		}

	   // ******************************************************************
	   // * func: XBController::~XBController
	   // ******************************************************************
	   public void dispose()
	   {
		   if (m_CurrentState == XBCtrlState.XBCTRL_STATE_CONFIG)
		   {
			   ConfigEnd();
		   }
		   else if (m_CurrentState == XBCtrlState.XBCTRL_STATE_LISTEN)
		   {
			   ListenEnd();
		   }
		   super.dispose();
	   }

		// ******************************************************************
		// * Registry Load/Save
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::Load
		// ******************************************************************
		public final void Load(String szRegistryKey)
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_NONE)
			{
				SetError("Invalid State", false);
				return;
			}

			// ******************************************************************
			// * Load Configuration from Registry
			// ******************************************************************
			{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: DWORD dwDisposition, dwType, dwSize;
				int dwDisposition;
				int dwType;
				int dwSize;
				HKEY hKey = new HKEY();

				if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegistryKey, 0, null, REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE, null, hKey, dwDisposition) == ERROR_SUCCESS)
				{
					int v = 0;

					// ******************************************************************
					// * Load Device Names
					// ******************************************************************
					{
						String szValueName = new String(new char[64]);

						for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
						{
							// default is a null string
							m_DeviceName[v][0] = (byte)'\0';

							szValueName = String.format("DeviceName 0x%.02X", v);

							dwType = REG_SZ;
							dwSize = 260;
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: RegQueryValueEx(hKey, szValueName, NULL, &dwType, (PBYTE)m_DeviceName[v], &dwSize);
							RegQueryValueEx(hKey, szValueName, null, dwType, (byte)m_DeviceName[v], dwSize);
						}
					}

					// ******************************************************************
					// * Load Object Configuration
					// ******************************************************************
					{
						String szValueName = new String(new char[64]);

						for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
						{
							// default object configuration
							m_ObjectConfig[v].dwDevice = -1;
							m_ObjectConfig[v].dwInfo = -1;
							m_ObjectConfig[v].dwFlags = 0;

							szValueName = String.format("Object : \"%s\"", GlobalMembers.m_DeviceNameLookup[v]);

							dwType = REG_BINARY;
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
							dwSize = sizeof(XBCtrlObjectCfg);
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: RegQueryValueEx(hKey, szValueName, NULL, &dwType, (PBYTE)&m_ObjectConfig[v], &dwSize);
							RegQueryValueEx(hKey, szValueName, null, dwType, (byte) m_ObjectConfig[v], dwSize);
						}
					}

					RegCloseKey(hKey);
				}
			}
		}

		// ******************************************************************
		// * func: XBController::Save
		// ******************************************************************
		public final void Save(String szRegistryKey)
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_NONE)
			{
				SetError("Invalid State", false);
				return;
			}

			// ******************************************************************
			// * Save Configuration to Registry
			// ******************************************************************
			{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: DWORD dwDisposition, dwType, dwSize;
				int dwDisposition;
				int dwType;
				int dwSize;
				HKEY hKey = new HKEY();

				if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegistryKey, 0, null, REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, null, hKey, dwDisposition) == ERROR_SUCCESS)
				{
					int v = 0;

					// ******************************************************************
					// * Save Device Names
					// ******************************************************************
					{
						String szValueName = new String(new char[64]);

						for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
						{
							szValueName = String.format("DeviceName 0x%.02X", v);

							dwType = REG_SZ;
							dwSize = 260;

							if (m_DeviceName[v][0] == '\0')
							{
								RegDeleteValue(hKey, szValueName);
							}
							else
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: RegSetValueEx(hKey, szValueName, NULL, dwType, (PBYTE)m_DeviceName[v], dwSize);
								RegSetValueEx(hKey, szValueName, null, dwType, (byte)m_DeviceName[v], dwSize);
							}
						}
					}

					// ******************************************************************
					// * Save Object Configuration
					// ******************************************************************
					{
						String szValueName = new String(new char[64]);

						for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
						{
							szValueName = String.format("Object : \"%s\"", GlobalMembers.m_DeviceNameLookup[v]);

							dwType = REG_BINARY;
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
							dwSize = sizeof(XBCtrlObjectCfg);

							if (m_ObjectConfig[v].dwDevice != -1)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: RegSetValueEx(hKey, szValueName, NULL, dwType, (PBYTE)&m_ObjectConfig[v], dwSize);
								RegSetValueEx(hKey, szValueName, null, dwType, (byte) m_ObjectConfig[v], dwSize);
							}
						}
					}

					RegCloseKey(hKey);
				}
			}
		}

		// ******************************************************************
		// * Configuration
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::ConfigBegin
		// ******************************************************************
		public final void ConfigBegin(HWND hwnd, XBCtrlObject object)
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_NONE)
			{
				SetError("Invalid State", false);
				return;
			}

			m_CurrentState = XBCtrlState.XBCTRL_STATE_CONFIG;

			DInputInit(hwnd);

			if (GetError() != 0)
			{
				return;
			}

			lPrevMouseX = -1;
			lPrevMouseY = -1;
			lPrevMouseZ = -1;

			CurConfigObject = object;

			return;
		}

		// ******************************************************************
		// * func: XBController::ConfigPoll
		// ******************************************************************
		public final boolean ConfigPoll(tangible.RefObject<String> szStatus)
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_CONFIG)
			{
				SetError("Invalid State", false);
				return false;
			}

			XTL.DIDEVICEINSTANCE DeviceInstance = new XTL.DIDEVICEINSTANCE();
			XTL.DIDEVICEOBJECTINSTANCE ObjectInstance = new XTL.DIDEVICEOBJECTINSTANCE();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			DeviceInstance.dwSize = sizeof(XTL.DIDEVICEINSTANCE);
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
			ObjectInstance.dwSize = sizeof(XTL.DIDEVICEOBJECTINSTANCE);

			// ******************************************************************
			// * Monitor for significant device state changes
			// ******************************************************************
			for (int v = m_dwInputDeviceCount - 1;v >= 0;v--)
			{
				// ******************************************************************
				// * Poll the current device
				// ******************************************************************
				{
					int hRet = m_InputDevice[v].m_Device.Poll();

					if (FAILED(hRet))
					{
						hRet = m_InputDevice[v].m_Device.Acquire();

						while (hRet == DIERR_INPUTLOST)
						{
							hRet = m_InputDevice[v].m_Device.Acquire();
						}
					}
				}

		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: DWORD dwHow = -1, dwFlags = m_InputDevice[v].m_Flags;
				int dwHow = -1;
				int dwFlags = m_InputDevice[v].m_Flags;

				// ******************************************************************
				// * Detect Joystick Input
				// ******************************************************************
				if ((m_InputDevice[v].m_Flags & 1 << 0) != 0)
				{
					XTL.DIJOYSTATE JoyState = new XTL.DIJOYSTATE();

					// ******************************************************************
					// * Get Joystick State
					// ******************************************************************
					{
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
						int hRet = m_InputDevice[v].m_Device.GetDeviceState(sizeof(XTL.DIJOYSTATE), JoyState);

						if (FAILED(hRet))
						{
							continue;
						}
					}

					dwFlags = 1 << 0;

					if (Math.abs(JoyState.lX) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lX));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lX));
						dwFlags |= (JoyState.lX > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else if (Math.abs(JoyState.lY) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lY));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lY));
						dwFlags |= (JoyState.lY > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else if (Math.abs(JoyState.lZ) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lZ));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lZ));
						dwFlags |= (JoyState.lZ > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else if (Math.abs(JoyState.lRx) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lRx));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lRx));
						dwFlags |= (JoyState.lRx > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else if (Math.abs(JoyState.lRy) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lRy));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lRy));
						dwFlags |= (JoyState.lRy > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else if (Math.abs(JoyState.lRz) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->lRz));
						dwHow = ((int) & (((XTL.DIJOYSTATE)0).lRz));
						dwFlags |= (JoyState.lRz > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
					}
					else
					{
						for (int b = 0;b < 2;b++)
						{
							if (Math.abs(JoyState.rglSlider[b]) > DefineConstants.DETECT_SENSITIVITY_JOYSTICK)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->rglSlider[b]));
								dwHow = ((int) & (((XTL.DIJOYSTATE)0).rglSlider[b]));
								dwFlags |= (JoyState.rglSlider[b] > 0) ? (1 << 3 | 1 << 5) : (1 << 3 | 1 << 6);
							}
						}
					}

					/* temporarily disabled
					if(dwHow == -1)
					{
					    for(int b=0;b<4;b++)
					    {
					        if(abs(JoyState.rgdwPOV[b]) > DETECT_SENSITIVITY_POV)
					        {
					            dwHow = FIELD_OFFSET(XTL::DIJOYSTATE, rgdwPOV[b]);
					        }
					    }
					}
					//*/

					if (dwHow == -1)
					{
						for (int b = 0;b < 32;b++)
						{
							if (JoyState.rgbButtons[b] > DefineConstants.DETECT_SENSITIVITY_BUTTON)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIJOYSTATE *)0)->rgbButtons[b]));
								dwHow = ((int) & (((XTL.DIJOYSTATE)0).rgbButtons[b]));
								dwFlags |= 1 << 4;
							}
						}
					}

					// ******************************************************************
					// * Retrieve Object Info
					// ******************************************************************
					if (dwHow != -1)
					{
						String szDirection = ((dwFlags & 1 << 3) != 0) ? ((dwFlags & 1 << 5) != 0) ? "Positive " : "Negative " : "";

						m_InputDevice[v].m_Device.GetDeviceInfo(DeviceInstance);

						m_InputDevice[v].m_Device.GetObjectInfo(ObjectInstance, dwHow, DIPH_BYOFFSET);

						Map(CurConfigObject, DeviceInstance.tszInstanceName, dwHow, dwFlags);

						System.out.printf("Cxbx: Detected %s%s on %s\n", szDirection, ObjectInstance.tszName, DeviceInstance.tszInstanceName, ObjectInstance.dwType);

						szStatus.argValue = String.format("Success: %s Mapped to '%s%s' on '%s'!", GlobalMembers.m_DeviceNameLookup[CurConfigObject.getValue()], szDirection, ObjectInstance.tszName, DeviceInstance.tszInstanceName);

						return true;
					}
				}
				// ******************************************************************
				// * Detect Keyboard Input
				// ******************************************************************
				else if (m_InputDevice[v].m_Flags & 1 << 1)
				{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: BYTE KeyState[256];
					byte[] KeyState = new byte[256];

					m_InputDevice[v].m_Device.GetDeviceState(256, KeyState);

					dwFlags = 1 << 1;

					// ******************************************************************
					// * Check for Keyboard State Change
					// ******************************************************************
					for (int r = 0;r < 256;r++)
					{
						if (KeyState[r] != 0)
						{
							dwHow = r;
							break;
						}
					}

					// ******************************************************************
					// * Check for Success
					// ******************************************************************
					if (dwHow != -1)
					{
						Map(CurConfigObject, "SysKeyboard", dwHow, dwFlags);

						System.out.printf("Cxbx: Detected Key %d on SysKeyboard\n", dwHow);

						szStatus.argValue = String.format("Success: %s Mapped to Key %d on SysKeyboard", GlobalMembers.m_DeviceNameLookup[CurConfigObject.getValue()], dwHow);

						return true;
					}
				}
				// ******************************************************************
				// * Detect Mouse Input
				// ******************************************************************
				else if (m_InputDevice[v].m_Flags & 1 << 2)
				{
					XTL.DIMOUSESTATE2 MouseState = new XTL.DIMOUSESTATE2();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
					m_InputDevice[v].m_Device.GetDeviceState(sizeof(MouseState), MouseState);

					dwFlags = 1 << 2;

					// ******************************************************************
					// * Detect Button State Change
					// ******************************************************************
					for (int r = 0;r < 4;r++)
					{
						// 0x80 is the mask for button push
						if ((MouseState.rgbButtons[r] & 0x80) != 0)
						{
							dwHow = r;
							dwFlags |= 1 << 7;
							break;
						}
					}
					// ******************************************************************
					// * Check for Success
					// ******************************************************************
					if (dwHow != -1)
					{
						Map(CurConfigObject, "SysMouse", dwHow, dwFlags);

						System.out.printf("Cxbx: Detected Button %d on SysMouse\n", dwHow);

						szStatus.argValue = String.format("Success: %s Mapped to Button %d on SysMouse", GlobalMembers.m_DeviceNameLookup[CurConfigObject.getValue()], dwHow);

						return true;
					}
					// ******************************************************************
					// * Check for Mouse Movement
					// ******************************************************************
					else
					{
						int lAbsDeltaX = 0;
						int lAbsDeltaY = 0;
						int lAbsDeltaZ = 0;
						int lDeltaX = 0;
						int lDeltaY = 0;
						int lDeltaZ = 0;

						if (lPrevMouseX == -1 || lPrevMouseY == -1 || lPrevMouseZ == -1)
						{
							lDeltaX = lDeltaY = lDeltaZ = 0;
						}
						else
						{
							lDeltaX = MouseState.lX - lPrevMouseX;
							lDeltaY = MouseState.lY - lPrevMouseY;
							lDeltaZ = MouseState.lZ - lPrevMouseZ;

							lAbsDeltaX = Math.abs(lDeltaX);
							lAbsDeltaY = Math.abs(lDeltaY);
							lAbsDeltaZ = Math.abs(lDeltaZ);
						}

						int lMax = (lAbsDeltaX > lAbsDeltaY) ? lAbsDeltaX : lAbsDeltaY;

						if (lAbsDeltaZ > lMax)
						{
							lMax = lAbsDeltaZ;
						}

						lPrevMouseX = MouseState.lX;
						lPrevMouseY = MouseState.lY;
						lPrevMouseZ = MouseState.lZ;

						if (lMax > DefineConstants.DETECT_SENSITIVITY_MOUSE)
						{
							dwFlags |= 1 << 3;

							if (lMax == lAbsDeltaX)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lX));
								dwHow = ((int) & (((XTL.DIMOUSESTATE)0).lX));
								dwFlags |= (lDeltaX > 0) ? 1 << 5 : 1 << 6;
							}
							else if (lMax == lAbsDeltaY)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lY));
								dwHow = ((int) & (((XTL.DIMOUSESTATE)0).lY));
								dwFlags |= (lDeltaY > 0) ? 1 << 5 : 1 << 6;
							}
							else if (lMax == lAbsDeltaZ)
							{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: dwHow = ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lZ));
								dwHow = ((int) & (((XTL.DIMOUSESTATE)0).lZ));
								dwFlags |= (lDeltaZ > 0) ? 1 << 5 : 1 << 6;
							}
						}

						// ******************************************************************
						// * Check for Success
						// ******************************************************************
						if (dwHow != -1)
						{
							String szDirection = ((dwFlags & 1 << 5) != 0) ? "Positive" : "Negative";
							String szObjName = "Unknown";

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
							ObjectInstance.dwSize = sizeof(ObjectInstance);

							if (m_InputDevice[v].m_Device.GetObjectInfo(ObjectInstance, dwHow, DIPH_BYOFFSET) == DI_OK)
							{
								szObjName = ObjectInstance.tszName;
							}

							Map(CurConfigObject, "SysMouse", dwHow, dwFlags);

							System.out.printf("Cxbx: Detected Movement on the %s%s on SysMouse\n", szDirection, szObjName);

							szStatus.argValue = String.format("Success: %s Mapped to %s%s on SysMouse", GlobalMembers.m_DeviceNameLookup[CurConfigObject.getValue()], szDirection, szObjName);

							return true;
						}
					}
				}
			}

			return false;
		}

		// ******************************************************************
		// * func: XBController::ConfigEnd
		// ******************************************************************
		public final void ConfigEnd()
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_CONFIG)
			{
				SetError("Invalid State", false);
				return;
			}

			DInputCleanup();

			m_CurrentState = XBCtrlState.XBCTRL_STATE_NONE;

			return;
		}

		// ******************************************************************
		// * Listening
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::ListenBegin
		// ******************************************************************
		public final void ListenBegin(HWND hwnd)
		{
			int v = 0;

			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_NONE)
			{
				SetError("Invalid State", false);
				return;
			}

			m_CurrentState = XBCtrlState.XBCTRL_STATE_LISTEN;

			DInputInit(hwnd);

			for (v = XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue() - 1;v >= m_dwInputDeviceCount;v--)
			{
				m_DeviceName[v][0] = (byte)'\0';
			}

			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (m_ObjectConfig[v].dwDevice >= m_dwInputDeviceCount)
				{
					System.out.printf("Warning: Device Mapped to %s was not found!\n", GlobalMembers.m_DeviceNameLookup[v]);
					m_ObjectConfig[v].dwDevice = -1;
				}
			}

			return;
		}

		// ******************************************************************
		// * func: XBController::ListenPoll
		// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: This was formerly a static local variable declaration (not allowed in Java):
		private int ListenPoll_lAccumX = 0;
//C++ TO JAVA CONVERTER NOTE: This was formerly a static local variable declaration (not allowed in Java):
		private int ListenPoll_lAccumY = 0;
//C++ TO JAVA CONVERTER NOTE: This was formerly a static local variable declaration (not allowed in Java):
		private int ListenPoll_lAccumZ = 0;
		public final void ListenPoll(_XINPUT_STATE Controller)
		{
			if (Controller == null)
			{
				return;
			}

			XTL.LPDIRECTINPUTDEVICE8 pDevice = null;

			int hRet = 0;
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: DWORD dwFlags=0;
			int dwFlags = 0;

			// ******************************************************************
			// * Default values necessary for axis
			// ******************************************************************
			Controller.Gamepad.sThumbLX = 0;
			Controller.Gamepad.sThumbLY = 0;
			Controller.Gamepad.sThumbRX = 0;
			Controller.Gamepad.sThumbRY = 0;

			// ******************************************************************
			// * Poll all devices
			// ******************************************************************
			for (int v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				int dwDevice = m_ObjectConfig[v].dwDevice;
				int dwFlags = m_ObjectConfig[v].dwFlags;
				int dwInfo = m_ObjectConfig[v].dwInfo;

				if (dwDevice == -1)
				{
					continue;
				}

				pDevice = m_InputDevice[dwDevice].m_Device;

				hRet = pDevice.Poll();

				if (FAILED(hRet))
				{
					hRet = pDevice.Acquire();

					while (hRet == DIERR_INPUTLOST)
					{
						hRet = pDevice.Acquire();
					}
				}

				short wValue = 0;

				// ******************************************************************
				// * Interpret PC Joystick Input
				// ******************************************************************
				if ((dwFlags & 1 << 0) != 0)
				{
					XTL.DIJOYSTATE JoyState = new XTL.DIJOYSTATE();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
					if (pDevice.GetDeviceState(sizeof(JoyState), JoyState) != DI_OK)
					{
						continue;
					}

					if ((dwFlags & 1 << 3) != 0)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: LONG *pdwAxis = (LONG*)((unsigned long)&JoyState + dwInfo);
//C++ TO JAVA CONVERTER TODO TASK: Java does not have an equivalent to pointers to value types:
//ORIGINAL LINE: int *pdwAxis = (int*)((uint)&JoyState + dwInfo);
						int pdwAxis = (int)((int) JoyState + dwInfo);
						wValue = (short)(pdwAxis);

						if ((dwFlags & 1 << 6) != 0)
						{
							if (wValue < 0)
							{
								wValue = Math.abs(wValue+1);
							}
							else
							{
								wValue = 0;
							}
						}
						else if (dwFlags & 1 << 5)
						{
							if (wValue < 0)
							{
								wValue = 0;
							}
						}
					}
					else if (dwFlags & 1 << 4)
					{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: BYTE *pbButton = (BYTE*)((unsigned long)&JoyState + dwInfo);
//C++ TO JAVA CONVERTER TODO TASK: Java does not have an equivalent to pointers to value types:
//ORIGINAL LINE: byte *pbButton = (byte*)((uint)&JoyState + dwInfo);
						byte pbButton = (byte)((int) JoyState + dwInfo);

						if ((pbButton & 0x80) != 0)
						{
							wValue = 32767;
						}
						else
						{
							wValue = 0;
						}
					}
				}
				// ******************************************************************
				// * Interpret PC KeyBoard Input
				// ******************************************************************
				else if (dwFlags & 1 << 1)
				{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: BYTE KeyboardState[256] = {0};
					byte[] KeyboardState = new byte[256];

					if (pDevice.GetDeviceState(Byte.SIZE, KeyboardState) != DI_OK)
					{
						continue;
					}

		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: BYTE bKey = KeyboardState[dwInfo];
					byte bKey = KeyboardState[dwInfo];

					if ((bKey & 0x80) != 0)
					{
						wValue = 32767;
					}
					else
					{
						wValue = 0;
					}
				}
				// ******************************************************************
				// * Interpret PC Mouse Input
				// ******************************************************************
				else if (dwFlags & 1 << 2)
				{
					XTL.DIMOUSESTATE2 MouseState = new XTL.DIMOUSESTATE2();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
					if (pDevice.GetDeviceState(sizeof(MouseState), MouseState) != DI_OK)
					{
						continue;
					}

					if ((dwFlags & 1 << 7) != 0)
					{
						if ((MouseState.rgbButtons[dwInfo] & 0x80) != 0)
						{
							wValue = 32767;
						}
						else
						{
							wValue = 0;
						}
					}
					else if (dwFlags & 1 << 3)
					{
//C++ TO JAVA CONVERTER NOTE: This static local variable declaration (not allowed in Java) has been moved just prior to the method:
//						static int lAccumX = 0;
//C++ TO JAVA CONVERTER NOTE: This static local variable declaration (not allowed in Java) has been moved just prior to the method:
//						static int lAccumY = 0;
//C++ TO JAVA CONVERTER NOTE: This static local variable declaration (not allowed in Java) has been moved just prior to the method:
//						static int lAccumZ = 0;

						ListenPoll_lAccumX += MouseState.lX * 300;
						ListenPoll_lAccumY += MouseState.lY * 300;
						ListenPoll_lAccumZ += MouseState.lZ * 300;

						if (ListenPoll_lAccumX > 32767)
						{
							ListenPoll_lAccumX = 32767;
						}
						else if (ListenPoll_lAccumX < -32768)
						{
							ListenPoll_lAccumX = -32768;
						}

						if (ListenPoll_lAccumY > 32767)
						{
							ListenPoll_lAccumY = 32767;
						}
						else if (ListenPoll_lAccumY < -32768)
						{
							ListenPoll_lAccumY = -32768;
						}

						if (ListenPoll_lAccumZ > 32767)
						{
							ListenPoll_lAccumZ = 32767;
						}
						else if (ListenPoll_lAccumZ < -32768)
						{
							ListenPoll_lAccumZ = -32768;
						}

		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: if(dwInfo == ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lX)))
						if (dwInfo == ((int) & (((XTL.DIMOUSESTATE)0).lX)))
						{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: wValue = (WORD)lAccumX;
							wValue = (short)ListenPoll_lAccumX;
						}
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: else if(dwInfo == ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lY)))
						else if (dwInfo == ((int) & (((XTL.DIMOUSESTATE)0).lY)))
						{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: wValue = (WORD)lAccumY;
							wValue = (short)ListenPoll_lAccumY;
						}
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: else if(dwInfo == ((ULONG)&(((XTL::DIMOUSESTATE *)0)->lZ)))
						else if (dwInfo == ((int) & (((XTL.DIMOUSESTATE)0).lZ)))
						{
		//C++ TO JAVA CONVERTER WARNING: Unsigned integer types have no direct equivalent in Java:
		//ORIGINAL LINE: wValue = (WORD)lAccumZ;
							wValue = (short)ListenPoll_lAccumZ;
						}

						if ((dwFlags & 1 << 6) != 0)
						{
							if (wValue < 0)
							{
								wValue = Math.abs(wValue+1);
							}
							else
							{
								wValue = 0;
							}
						}
						else if (dwFlags & 1 << 5)
						{
							if (wValue < 0)
							{
								wValue = 0;
							}
						}
					}
				}

				// ******************************************************************
				// * Map Xbox Joystick Input
				// ******************************************************************
				if (v >= XBCtrlObject.XBCTRL_OBJECT_LTHUMBPOSX.getValue() != 0 && v <= XBCtrlObject.XBCTRL_OBJECT_RTHUMB.getValue())
				{
					switch (v)
					{
						case XBCTRL_OBJECT_LTHUMBPOSY:
							Controller.Gamepad.sThumbLY += wValue;
							break;
						case XBCTRL_OBJECT_LTHUMBNEGY:
							Controller.Gamepad.sThumbLY -= wValue;
							break;
						case XBCTRL_OBJECT_RTHUMBPOSY:
							Controller.Gamepad.sThumbRY += wValue;
							break;
						case XBCTRL_OBJECT_RTHUMBNEGY:
							Controller.Gamepad.sThumbRY -= wValue;
							break;
						case XBCTRL_OBJECT_LTHUMBPOSX:
							Controller.Gamepad.sThumbLX += wValue;
							break;
						case XBCTRL_OBJECT_LTHUMBNEGX:
							Controller.Gamepad.sThumbLX -= wValue;
							break;
						case XBCTRL_OBJECT_RTHUMBPOSX:
							Controller.Gamepad.sThumbRX += wValue;
							break;
						case XBCTRL_OBJECT_RTHUMBNEGX:
							Controller.Gamepad.sThumbRX -= wValue;
							break;
						case XBCTRL_OBJECT_A:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_A] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_B:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_B] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_X:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_X] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_Y:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_Y] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_WHITE:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_WHITE] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_BLACK:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_BLACK] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_LTRIGGER:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_LEFT_TRIGGER] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_RTRIGGER:
							Controller.Gamepad.bAnalogButtons[DefineConstants.XINPUT_GAMEPAD_RIGHT_TRIGGER] = (wValue / 128);
							break;
						case XBCTRL_OBJECT_DPADUP:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_DPAD_UP;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_DPAD_UP;
							}
							break;
						case XBCTRL_OBJECT_DPADDOWN:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_DPAD_DOWN;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_DPAD_DOWN;
							}
							break;
						case XBCTRL_OBJECT_DPADLEFT:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_DPAD_LEFT;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_DPAD_LEFT;
							}
							break;
						case XBCTRL_OBJECT_DPADRIGHT:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_DPAD_RIGHT;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_DPAD_RIGHT;
							}
							break;
						case XBCTRL_OBJECT_BACK:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_BACK;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_BACK;
							}
							break;
						case XBCTRL_OBJECT_START:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_START;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_START;
							}
							break;
						case XBCTRL_OBJECT_LTHUMB:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_LEFT_THUMB;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_LEFT_THUMB;
							}
							break;
						case XBCTRL_OBJECT_RTHUMB:
							if (wValue > 0)
							{
								Controller.Gamepad.wButtons |= DefineConstants.XINPUT_GAMEPAD_RIGHT_THUMB;
							}
							else
							{
								Controller.Gamepad.wButtons &= ~DefineConstants.XINPUT_GAMEPAD_RIGHT_THUMB;
							}
							break;
					}
				}
			}

			return;
		}

		// ******************************************************************
		// * func: XBController::ListenEnd
		// ******************************************************************
		public final void ListenEnd()
		{
			if (m_CurrentState != XBCtrlState.XBCTRL_STATE_LISTEN)
			{
				SetError("Invalid State", false);
				return;
			}

			DInputCleanup();

			m_CurrentState = XBCtrlState.XBCTRL_STATE_NONE;

			return;
		}

		// ******************************************************************
		// * DirectInput Init / Cleanup
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::DInputInit
		// ******************************************************************
		public final void DInputInit(HWND hwnd)
		{
			m_dwInputDeviceCount = null;

			// ******************************************************************
			// * Create DirectInput Object
			// ******************************************************************
			{
				int hRet = XTL.DirectInput8Create(GetModuleHandle(null), DefineConstants.DIRECTINPUT_VERSION, XTL.IID_IDirectInput8, (Object) m_pDirectInput8, null);

				if (FAILED(hRet))
				{
					SetError("Could not initialized DirectInput8", true);
					return;
				}
			}

			// ******************************************************************
			// * Create all the devices available (well...most of them)
			// ******************************************************************
			if (m_pDirectInput8 != 0)
			{
				int hRet = m_pDirectInput8.EnumDevices(DI8DEVCLASS_GAMECTRL, WrapEnumGameCtrlCallback, this, DIEDFL_ATTACHEDONLY);

				if (m_CurrentState == XBCtrlState.XBCTRL_STATE_CONFIG || DeviceIsUsed("SysKeyboard"))
				{
					hRet = m_pDirectInput8.CreateDevice(XTL.GUID_SysKeyboard, m_InputDevice[m_dwInputDeviceCount].m_Device, null);

					if (!FAILED(hRet))
					{
						m_InputDevice[m_dwInputDeviceCount].m_Flags = 1 << 1;

						m_InputDevice[m_dwInputDeviceCount++].m_Device.SetDataFormat(XTL.c_dfDIKeyboard);
					}

					if (m_CurrentState == XBCtrlState.XBCTRL_STATE_LISTEN)
					{
						ReorderObjects("SysKeyboard", m_dwInputDeviceCount - 1);
					}
				}

				if (m_CurrentState == XBCtrlState.XBCTRL_STATE_CONFIG || DeviceIsUsed("SysMouse"))
				{
					hRet = m_pDirectInput8.CreateDevice(XTL.GUID_SysMouse, m_InputDevice[m_dwInputDeviceCount].m_Device, null);

					if (!FAILED(hRet))
					{
						m_InputDevice[m_dwInputDeviceCount].m_Flags = 1 << 2;

						m_InputDevice[m_dwInputDeviceCount++].m_Device.SetDataFormat(XTL.c_dfDIMouse2);
					}

					if (m_CurrentState == XBCtrlState.XBCTRL_STATE_LISTEN)
					{
						ReorderObjects("SysMouse", m_dwInputDeviceCount - 1);
					}
				}
			}

			// ******************************************************************
			// * Enumerate Controller objects
			// ******************************************************************
			for (m_dwCurObject = 0;m_dwCurObject < m_dwInputDeviceCount;m_dwCurObject++)
			{
				m_InputDevice[m_dwCurObject].m_Device.EnumObjects(WrapEnumObjectsCallback, this, DIDFT_ALL);
			}

			// ******************************************************************
			// * Set cooperative level and acquire
			// ******************************************************************
			{
				for (int v = m_dwInputDeviceCount - 1;v >= 0;v--)
				{
					m_InputDevice[v].m_Device.SetCooperativeLevel(hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
					m_InputDevice[v].m_Device.Acquire();

					int hRet = m_InputDevice[v].m_Device.Poll();

					if (FAILED(hRet))
					{
						hRet = m_InputDevice[v].m_Device.Acquire();

						while (hRet == DIERR_INPUTLOST)
						{
							hRet = m_InputDevice[v].m_Device.Acquire();
						}

						if (hRet != DIERR_INPUTLOST)
						{
							break;
						}
					}
				}
			}
		}

		// ******************************************************************
		// * func: XBController::DInputCleanup
		// ******************************************************************
		public final void DInputCleanup()
		{
			for (int v = m_dwInputDeviceCount - 1;v >= 0;v--)
			{
				m_InputDevice[v].m_Device.Unacquire();
				m_InputDevice[v].m_Device.Release();
				m_InputDevice[v].m_Device = 0;
			}

			m_dwInputDeviceCount = 0;

			if (m_pDirectInput8 != 0)
			{
				m_pDirectInput8.Release();
				m_pDirectInput8 = 0;
			}

			return;
		}

		// ******************************************************************
		// * Check if a device is currently in the configuration
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::DeviceIsUsed
		// ******************************************************************
		public final boolean DeviceIsUsed(String szDeviceName)
		{
			for (int v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (m_DeviceName[v][0] != '\0')
				{
					if (strncmp(m_DeviceName[v], szDeviceName, 255) == 0)
					{
						return true;
					}
				}
			}

			return false;
		}

		// ******************************************************************
		// * Input Device Name Lookup Table
		// ******************************************************************
		public final String[] m_DeviceNameLookup = new String[XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue()];

		// ******************************************************************
		// * Object Mapping
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::Map
		// ******************************************************************
		private void Map(XBCtrlObject object, String szDeviceName, int dwInfo, int dwFlags)
		{
			// Initialize InputMapping instance
			m_ObjectConfig[object.getValue()].dwDevice = Insert(szDeviceName);
			m_ObjectConfig[object.getValue()].dwInfo = dwInfo;
			m_ObjectConfig[object.getValue()].dwFlags = dwFlags;

			// Purge unused device slots
			for (int v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				boolean inuse = false;

				for (int r = 0;r < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();r++)
				{
					if (m_ObjectConfig[r].dwDevice == v)
					{
						inuse = true;
					}
				}

				if (!inuse)
				{
					m_DeviceName[v][0] = (byte)'\0';
				}
			}
		}

		// ******************************************************************
		// * Find the look-up value for a DeviceName (creating if needed)
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::Insert
		// ******************************************************************
		private int Insert(String szDeviceName)
		{
			int v = 0;

			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (strcmp(m_DeviceName[v], szDeviceName) == 0)
				{
					return v;
				}
			}

			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (m_DeviceName[v][0] == '\0')
				{
					m_DeviceName[v] = szDeviceName.substring(0, 255);

					return v;
				}
			}

			JOptionPane.showConfirmDialog(null, "Unexpected Circumstance (Too Many Controller Devices)! Please contact caustik!", "Cxbx", JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE);

			ExitProcess(1);

			return 0;
		}

		// ******************************************************************
		// * Update the object lookup offsets for a device
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::ReorderObjects
		// ******************************************************************
		private void ReorderObjects(String szDeviceName, int pos)
		{
			int old = -1;
			int v = 0;

			// locate old device name position
			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (strcmp(m_DeviceName[v], szDeviceName) == 0)
				{
					old = v;
					break;
				}
			}

			// Swap names, if necessary
			if (old != pos)
			{
				m_DeviceName[old] = m_DeviceName[pos];
				m_DeviceName[pos] = szDeviceName;
			}

			// Update all old values
			for (v = 0;v < XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue();v++)
			{
				if (m_ObjectConfig[v].dwDevice == old)
				{
					m_ObjectConfig[v].dwDevice = pos;
				}
				else if (m_ObjectConfig[v].dwDevice == pos)
				{
					m_ObjectConfig[v].dwDevice = old;
				}
			}

			return;
		}

		// ******************************************************************
		// * Controller and Objects Enumeration
		// ******************************************************************

		// ******************************************************************
		// * func: XBController::EnumGameCtrlCallback
		// ******************************************************************
		private int EnumGameCtrlCallback(XTL.LPCDIDEVICEINSTANCE lpddi)
		{
			if (m_CurrentState == XBCtrlState.XBCTRL_STATE_LISTEN && !DeviceIsUsed(lpddi.tszInstanceName))
			{
				return DIENUM_CONTINUE;
			}

			int hRet = m_pDirectInput8.CreateDevice(lpddi.guidInstance, m_InputDevice[m_dwInputDeviceCount].m_Device, null);

			if (!FAILED(hRet))
			{
				m_InputDevice[m_dwInputDeviceCount].m_Flags = 1 << 0;

				m_InputDevice[m_dwInputDeviceCount++].m_Device.SetDataFormat(XTL.c_dfDIJoystick);

				if (m_CurrentState == XBCtrlState.XBCTRL_STATE_LISTEN)
				{
					ReorderObjects(lpddi.tszInstanceName, m_dwInputDeviceCount - 1);
				}
			}

			return DIENUM_CONTINUE;
		}

		// ******************************************************************
		// * func: XBController::EnumObjectsCallback
		// ******************************************************************
		private int EnumObjectsCallback(XTL.LPCDIDEVICEOBJECTINSTANCE lpddoi)
		{
			if ((lpddoi.dwType & DIDFT_AXIS) != 0)
			{
				XTL.DIPROPRANGE diprg = new XTL.DIPROPRANGE();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
				diprg.diph.dwSize = sizeof(XTL.DIPROPRANGE);
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
				diprg.diph.dwHeaderSize = sizeof(XTL.DIPROPHEADER);
				diprg.diph.dwHow = DIPH_BYID;
				diprg.diph.dwObj = lpddoi.dwType;
				diprg.lMin = 0 - 32768;
				diprg.lMax = 0 + 32767;

				int hRet = m_InputDevice[m_dwCurObject].m_Device.SetProperty(DIPROP_RANGE, diprg.diph);

				if (FAILED(hRet))
				{
					if (hRet == E_NOTIMPL)
					{
						return DIENUM_CONTINUE;
					}
					else
					{
						return DIENUM_STOP;
					}
				}
			}
			else if (lpddoi.dwType & DIDFT_BUTTON)
			{
				XTL.DIPROPRANGE diprg = new XTL.DIPROPRANGE();

//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
				diprg.diph.dwSize = sizeof(XTL.DIPROPRANGE);
//C++ TO JAVA CONVERTER TODO TASK: There is no Java equivalent to 'sizeof':
				diprg.diph.dwHeaderSize = sizeof(XTL.DIPROPHEADER);
				diprg.diph.dwHow = DIPH_BYID;
				diprg.diph.dwObj = lpddoi.dwType;
				diprg.lMin = 0;
				diprg.lMax = 255;

				int hRet = m_InputDevice[m_dwCurObject].m_Device.SetProperty(DIPROP_RANGE, diprg.diph);

				if (FAILED(hRet))
				{
					if (hRet == E_NOTIMPL)
					{
						return DIENUM_CONTINUE;
					}
					else
					{
						return DIENUM_STOP;
					}
				}
			}

			return DIENUM_CONTINUE;
		}

		// ******************************************************************
		// * Wrapper Function for Enumeration
		// ******************************************************************

		// ******************************************************************
		// * func: WrapEnumGameCtrlCallback
		// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: CALLBACK is not available in Java:
//ORIGINAL LINE: int CALLBACK WrapEnumGameCtrlCallback(XTL::LPCDIDEVICEINSTANCE lpddi, Object*  pvRef)
		private int WrapEnumGameCtrlCallback(XTL.LPCDIDEVICEINSTANCE lpddi, Object pvRef)
		{
			XBController context = (XBController)pvRef;

			return context.EnumGameCtrlCallback(lpddi);
		}

		// ******************************************************************
		// * func: WrapEnumObjectsCallback
		// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: CALLBACK is not available in Java:
//ORIGINAL LINE: int CALLBACK WrapEnumObjectsCallback(XTL::LPCDIDEVICEOBJECTINSTANCE lpddoi, Object*  pvRef)
		private int WrapEnumObjectsCallback(XTL.LPCDIDEVICEOBJECTINSTANCE lpddoi, Object pvRef)
		{
			XBController context = (XBController)pvRef;

			return context.EnumObjectsCallback(lpddoi);
		}

		// ******************************************************************
		// * Device Names
		// ******************************************************************
		private byte[][] m_DeviceName = new byte[XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue()][260];

		// ******************************************************************
		// * Object Configuration
		// ******************************************************************
		private XBCtrlObjectCfg[] m_ObjectConfig = tangible.Arrays.initializeWithDefaultXBCtrlObjectCfgInstances(XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue());

		// ******************************************************************
		// * DirectInput
		// ******************************************************************
		private XTL.LPDIRECTINPUT8 m_pDirectInput8 = new XTL.LPDIRECTINPUT8();

		// ******************************************************************
		// * DirectInput Devices
		// ******************************************************************
		private static class InputDevice
		{
			public XTL.LPDIRECTINPUTDEVICE8 m_Device = new XTL.LPDIRECTINPUTDEVICE8();
			public int m_Flags;
		}
		private InputDevice[] m_InputDevice = tangible.Arrays.initializeWithDefaultInputDeviceInstances(XBCtrlObject.XBCTRL_OBJECT_COUNT.getValue());

		// ******************************************************************
		// * Current State
		// ******************************************************************
		private XBCtrlState m_CurrentState;

		// ******************************************************************
		// * Config State Variables
		// ******************************************************************
		private int lPrevMouseX;
		private int lPrevMouseY;
		private int lPrevMouseZ;
		private XBCtrlObject CurConfigObject;

		// ******************************************************************
		// * Etc State Variables
		// ******************************************************************
		private int m_dwInputDeviceCount;
		private int m_dwCurObject;

}

// ******************************************************************
// * Device Flags
// ******************************************************************
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_JOYSTICK (1 << 0)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_KEYBOARD (1 << 1)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_MOUSE (1 << 2)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_AXIS (1 << 3)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_BUTTON (1 << 4)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_POSITIVE (1 << 5)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_NEGATIVE (1 << 6)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_MOUSE_CLICK (1 << 7)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_MOUSE_LX (1 << 8)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_MOUSE_LY (1 << 9)
//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
//ORIGINAL LINE: #define DEVICE_FLAG_MOUSE_LZ (1 << 10)

// ******************************************************************
// * Detection Sensitivity
// ******************************************************************

// ******************************************************************
// * DirectInput Enumeration Types
// ******************************************************************
public enum XBCtrlState
{
	XBCTRL_STATE_NONE(0),
	XBCTRL_STATE_CONFIG(1),
	XBCTRL_STATE_LISTEN(2);

	private int intValue;
	private static java.util.HashMap<Integer, XBCtrlState> mappings;
	private static java.util.HashMap<Integer, XBCtrlState> getMappings()
	{
		if (mappings == null)
		{
			synchronized (XBCtrlState.class)
			{
				if (mappings == null)
				{
					mappings = new java.util.HashMap<Integer, XBCtrlState>();
				}
			}
		}
		return mappings;
	}

	private XBCtrlState(int value)
	{
		intValue = value;
		getMappings().put(value, this);
	}

	public int getValue()
	{
		return intValue;
	}

	public static XBCtrlState forValue(int value)
	{
		return getMappings().get(value);
	}
}


public class GlobalMembers
{
	// This is ridiculous
	//C++ TO JAVA CONVERTER NOTE: The following #define macro was replaced in-line:
	//ORIGINAL LINE: #define FIELD_OFFSET(type,field) ((ULONG)&(((type *)0)->field))

	// ******************************************************************
	// * Input Device Name Lookup Table
	// ******************************************************************
	public static String[] m_DeviceNameLookup = {"LThumbPosX", "LThumbNegX", "LThumbPosY", "LThumbNegY", "RThumbPosX", "RThumbNegX", "RThumbPosY", "RThumbNegY", "X", "Y", "A", "B", "White", "Black", "LTrigger", "RTrigger", "DPadUp", "DPadDown", "DPadLeft", "DPadRight", "Back", "Start", "LThumb", "RThumb"};
}

final class DefineConstants
{
//C++ TO JAVA CONVERTER TODO TASK: The following #define constant was defined in different ways:
	public static final String _CXBX_VERSION = "0.8.0";
	public static final String _CXBX_VERSION = "0.8.0-Trace";
	public static final int MAXIMUM_XBOX_THREADS = 256;
	public static final int XINPUT_DEVSUBTYPE_GC_GAMEPAD = 0x01;
	public static final int XINPUT_DEVSUBTYPE_GC_GAMEPAD_ALT = 0x02;
	public static final int XINPUT_DEVSUBTYPE_GC_WHEEL = 0x10;
	public static final int XINPUT_DEVSUBTYPE_GC_ARCADE_STICK = 0x20;
	public static final int XINPUT_DEVSUBTYPE_GC_DIGITAL_ARCADE_STICK = 0x21;
	public static final int XINPUT_DEVSUBTYPE_GC_FLIGHT_STICK = 0x30;
	public static final int XINPUT_DEVSUBTYPE_GC_SNOWBOARD = 0x40;
	public static final int XCALCSIG_SIGNATURE_SIZE = 20;
	public static final int XCALCSIG_FLAG_NON_ROAMABLE = 0x00000001;
	public static final int DIRECT3D_VERSION = 0x0800;
	public static final int X_PIXELSHADER_FAKE_HANDLE = 0xDEADBEEF;
	public static final int X_D3DCOMMON_REFCOUNT_MASK = 0x0000FFFF;
	public static final int X_D3DCOMMON_TYPE_MASK = 0x00070000;
	public static final int X_D3DCOMMON_TYPE_SHIFT = 16;
	public static final int X_D3DCOMMON_TYPE_VERTEXBUFFER = 0x00000000;
	public static final int X_D3DCOMMON_TYPE_INDEXBUFFER = 0x00010000;
	public static final int X_D3DCOMMON_TYPE_PUSHBUFFER = 0x00020000;
	public static final int X_D3DCOMMON_TYPE_PALETTE = 0x00030000;
	public static final int X_D3DCOMMON_TYPE_TEXTURE = 0x00040000;
	public static final int X_D3DCOMMON_TYPE_SURFACE = 0x00050000;
	public static final int X_D3DCOMMON_TYPE_FIXUP = 0x00060000;
	public static final int X_D3DCOMMON_INTREFCOUNT_MASK = 0x00780000;
	public static final int X_D3DCOMMON_INTREFCOUNT_SHIFT = 19;
	public static final int X_D3DCOMMON_D3DCREATED = 0x01000000;
	public static final int X_D3DCOMMON_ISLOCKED = 0x02000010; // Surface is currently locked (potential unswizzle candidate)
	public static final int X_D3DCOMMON_UNUSED_MASK = 0xFE000000;
	public static final int X_D3DCOMMON_UNUSED_SHIFT = 25;
	public static final int X_D3DRESOURCE_DATA_FLAG_SPECIAL = 0xFFFF0000;
	public static final int X_D3DRESOURCE_DATA_FLAG_SURFACE = 0x00000001; // Backbuffer surface, etc
	public static final int X_D3DRESOURCE_DATA_FLAG_YUVSURF = 0x00000002; // YUV memory surface
	public static final int X_D3DRESOURCE_DATA_FLAG_D3DREND = 0x00000004; // D3D Render Target
	public static final int X_D3DRESOURCE_DATA_FLAG_D3DSTEN = 0x00000008; // D3D Stencil Surface
	public static final int X_D3DRESOURCE_LOCK_FLAG_NOSIZE = 0xEFFFFFFF;
	public static final int X_D3DFORMAT_RESERVED1_MASK = 0x00000003; // Must be zero
	public static final int X_D3DFORMAT_DMACHANNEL_MASK = 0x00000003;
	public static final int X_D3DFORMAT_DMACHANNEL_A = 0x00000001; // DMA channel A - the default for all system memory
	public static final int X_D3DFORMAT_DMACHANNEL_B = 0x00000002; // DMA channel B - unused
	public static final int X_D3DFORMAT_CUBEMAP = 0x00000004; // Set if the texture if a cube map
	public static final int X_D3DFORMAT_BORDERSOURCE_COLOR = 0x00000008;
	public static final int X_D3DFORMAT_DIMENSION_MASK = 0x000000F0; // # of dimensions
	public static final int X_D3DFORMAT_DIMENSION_SHIFT = 4;
	public static final int X_D3DFORMAT_FORMAT_MASK = 0x0000FF00;
	public static final int X_D3DFORMAT_FORMAT_SHIFT = 8;
	public static final int X_D3DFORMAT_MIPMAP_MASK = 0x000F0000;
	public static final int X_D3DFORMAT_MIPMAP_SHIFT = 16;
	public static final int X_D3DFORMAT_USIZE_MASK = 0x00F00000; // Log 2 of the U size of the base texture
	public static final int X_D3DFORMAT_USIZE_SHIFT = 20;
	public static final int X_D3DFORMAT_VSIZE_MASK = 0x0F000000; // Log 2 of the V size of the base texture
	public static final int X_D3DFORMAT_VSIZE_SHIFT = 24;
	public static final int X_D3DFORMAT_PSIZE_MASK = 0xF0000000; // Log 2 of the P size of the base texture
	public static final int X_D3DFORMAT_PSIZE_SHIFT = 28;
	public static final int X_D3DSIZE_WIDTH_MASK = 0x00000FFF; // Width  (Texels - 1)
	public static final int X_D3DSIZE_HEIGHT_MASK = 0x00FFF000; // Height (Texels - 1)
	public static final int X_D3DSIZE_HEIGHT_SHIFT = 12;
	public static final int X_D3DSIZE_PITCH_MASK = 0xFF000000; // Pitch / 64 - 1
	public static final int X_D3DSIZE_PITCH_SHIFT = 24;
	public static final int X_D3DRS_UNK = 0x7fffffff;
	public static final int X_D3DTSS_UNK = 0x7fffffff;
	public static final int X_VSCM_96 = 0;
	public static final int X_VSCM_192 = 1;
	public static final int X_VSCM_192FIXEDPIPELINE = 2;
	public static final int X_VSCM_NONERESERVED = 16;
	public static final int X_VST_NORMAL = 1;
	public static final int X_VST_READWRITE = 2;
	public static final int X_VST_STATE = 3;
	public static final int XINPUT_SETSTATE_LATENCY = 4;
	public static final int XINPUT_SETSTATE_SLOTS = 16;
	public static final int XINPUT_HANDLE_SLOTS = 4;
	public static final int DIRECTDRAW_VERSION = 0x0700;
	public static final int X_D3DRSSE_UNK = 0x7fffffff;
	public static final int MAX_NBR_STREAMS = 16;
	public static final int VSH_INSTRUCTION_SIZE = 4;
	public static final int DIRECTINPUT_VERSION = 0x0800;
	public static final int XINPUT_GAMEPAD_A = 0;
	public static final int XINPUT_GAMEPAD_B = 1;
	public static final int XINPUT_GAMEPAD_X = 2;
	public static final int XINPUT_GAMEPAD_Y = 3;
	public static final int XINPUT_GAMEPAD_BLACK = 4;
	public static final int XINPUT_GAMEPAD_WHITE = 5;
	public static final int XINPUT_GAMEPAD_LEFT_TRIGGER = 6;
	public static final int XINPUT_GAMEPAD_RIGHT_TRIGGER = 7;
	public static final int XINPUT_GAMEPAD_DPAD_UP = 0x00000001;
	public static final int XINPUT_GAMEPAD_DPAD_DOWN = 0x00000002;
	public static final int XINPUT_GAMEPAD_DPAD_LEFT = 0x00000004;
	public static final int XINPUT_GAMEPAD_DPAD_RIGHT = 0x00000008;
	public static final int XINPUT_GAMEPAD_START = 0x00000010;
	public static final int XINPUT_GAMEPAD_BACK = 0x00000020;
	public static final int XINPUT_GAMEPAD_LEFT_THUMB = 0x00000040;
	public static final int XINPUT_GAMEPAD_RIGHT_THUMB = 0x00000080;
	public static final int DSB_FLAG_ADPCM = 0x00000001;
	public static final int DETECT_SENSITIVITY_JOYSTICK = 25000;
	public static final int DETECT_SENSITIVITY_BUTTON = 0;
	public static final int DETECT_SENSITIVITY_MOUSE = 5;
	public static final int DETECT_SENSITIVITY_POV = 50000;
}

package tangible;

//----------------------------------------------------------------------------------------
//	Copyright © 2006 - 2016 Tangible Software Solutions Inc.
//	This class can be used by anyone provided that the copyright notice remains intact.
//
//	This class provides the ability to initialize array elements with the default
//	constructions for the array type.
//----------------------------------------------------------------------------------------
public final class Arrays
{
	public static XBCtrlObjectCfg[] initializeWithDefaultXBCtrlObjectCfgInstances(int length)
	{
		XBCtrlObjectCfg[] array = new XBCtrlObjectCfg[length];
		for (int i = 0; i < length; i++)
		{
			array[i] = new XBCtrlObjectCfg();
		}
		return array;
	}

	public static InputDevice[] initializeWithDefaultInputDeviceInstances(int length)
	{
		InputDevice[] array = new InputDevice[length];
		for (int i = 0; i < length; i++)
		{
			array[i] = new InputDevice();
		}
		return array;
	}
}

package tangible;

//----------------------------------------------------------------------------------------
//	Copyright © 2006 - 2016 Tangible Software Solutions Inc.
//	This class can be used by anyone provided that the copyright notice remains intact.
//
//	This class is used to simulate the ability to pass arguments by reference in Java.
//----------------------------------------------------------------------------------------
public final class RefObject<T>
{
	public T argValue;
	public RefObject(T refArg)
	{
		argValue = refArg;
	}
}